// ********************
// * MaxHumans Plugin *
// ********************
// version 1.14
// by Pierre-Marie Baty <pm@bots-united.com>
// and AustinBotServer <austinbotserver@mindspring.com>

// LAST MODIFIED BY PM (change this line when you modify the CPP file)

#include <extdll.h>
#include <dllapi.h>
#include <h_export.h>
#include <meta_api.h>


// admin players WONId file
#define FILE_ADMINPLAYERS "cstrike/serveradmins.cfg"

// client messages
#define MESSAGE_REFUSE_CLIENTCONNECT "The maximal number of humans allowed on this server has been reached. Please come back later!\n"
#define MESSAGE_WARNING_COULDNOTKICKBOT "WARNING: There seems to be no bots on this server. No one has been kicked to let you enter.\n"

enginefuncs_t g_engfuncs;
globalvars_t  *gpGlobals;
DLL_FUNCTIONS gFunctionTable;
cvar_t maxhumans = {"maxhumans", "32", FCVAR_EXTDLL};
cvar_t allowadminson = {"allowadminson", "1", FCVAR_EXTDLL};
cvar_t addbot_command = {"addbot_command", "\"addbot 10 1 4\"", FCVAR_EXTDLL};
int player_count = 0;
int bot_count = 0;
int human_count = 0;
bool add_bot_now = FALSE;

void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals);
BOOL ClientConnect (edict_t *pEntity, const char *pszName, const char *pszAddress, char szRejectReason[128]);
void ClientDisconnect (edict_t *pEntity);
void StartFrame (void);
edict_t *CreateFakeClient (const char *netname);
bool IsAdministrator (edict_t *pPlayer);
void GetGameDir (char *game_dir);


// START of Metamod stuff

enginefuncs_t meta_engfuncs;
gamedll_funcs_t *gpGamedllFuncs;
mutil_funcs_t *gpMetaUtilFuncs;
meta_globals_t *gpMetaGlobals;

META_FUNCTIONS gMetaFunctionTable =
{
   NULL, // pfnGetEntityAPI()
   NULL, // pfnGetEntityAPI_Post()
   GetEntityAPI2, // pfnGetEntityAPI2()
   NULL, // pfnGetEntityAPI2_Post()
   NULL, // pfnGetNewDLLFunctions()
   NULL, // pfnGetNewDLLFunctions_Post()
   GetEngineFunctions, // pfnGetEngineFunctions()
   NULL, // pfnGetEngineFunctions_Post()
};

plugin_info_t Plugin_info = {
   META_INTERFACE_VERSION, // interface version
   "MaxHumans", // plugin name
   "1.14", // plugin version
   "30/05/2003", // date of creation
   "Pierre-Marie Baty <pm@bots-united.com> & AustinBotServer <austinbotserver@mindspring.com>", // plugin author
   "http://racc.bots-united.com", // plugin URL
   "MAXHUMANS", // plugin logtag
   PT_CHANGELEVEL, // when loadable
   PT_ANYTIME, // when unloadable
};


C_DLLEXPORT int Meta_Query (char *ifvers, plugin_info_t **pPlugInfo, mutil_funcs_t *pMetaUtilFuncs)
{
   // this function is the first function ever called by metamod in the plugin DLL. Its purpose
   // is for metamod to retrieve basic information about the plugin, such as its meta-interface
   // version, for ensuring compatibility with the current version of the running metamod.

   // keep track of the pointers to metamod function tables metamod gives us
   gpMetaUtilFuncs = pMetaUtilFuncs;
   *pPlugInfo = &Plugin_info;

   // check for interface version compatibility
   if (strcmp (ifvers, Plugin_info.ifvers) != 0)
   {
      int mmajor = 0, mminor = 0, pmajor = 0, pminor = 0;

      LOG_CONSOLE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);
      LOG_MESSAGE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);

      // if plugin has later interface version, it's incompatible (update metamod)
      sscanf (ifvers, "%d:%d", &mmajor, &mminor);
      sscanf (META_INTERFACE_VERSION, "%d:%d", &pmajor, &pminor);

      if ((pmajor > mmajor) || ((pmajor == mmajor) && (pminor > mminor)))
      {
         LOG_CONSOLE (PLID, "metamod version is too old for this plugin; update metamod");
         LOG_ERROR (PLID, "metamod version is too old for this plugin; update metamod");
         return (FALSE);
      }

      // if plugin has older major interface version, it's incompatible (update plugin)
      else if (pmajor < mmajor)
      {
         LOG_CONSOLE (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         LOG_ERROR (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         return (FALSE);
      }
   }

   return (TRUE); // tell metamod this plugin looks safe
}


C_DLLEXPORT int Meta_Attach (PLUG_LOADTIME now, META_FUNCTIONS *pFunctionTable, meta_globals_t *pMGlobals, gamedll_funcs_t *pGamedllFuncs)
{
   // this function is called when metamod attempts to load the plugin. Since it's the place
   // where we can tell if the plugin will be allowed to run or not, we wait until here to make
   // our initialization stuff, like registering CVARs and dedicated server commands.

   char dirname[256];

   // are we allowed to load this plugin now ?
   if (now > Plugin_info.loadable)
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }

   // keep track of the pointers to engine function tables metamod gives us
   gpMetaGlobals = pMGlobals;
   memcpy (pFunctionTable, &gMetaFunctionTable, sizeof (META_FUNCTIONS));
   gpGamedllFuncs = pGamedllFuncs;

   // get the MOD directory name (full path)
   GetGameDir (dirname);

   // is it NOT the Counter-Strike directory ?
   if (strcmp (dirname, "cstrike") != 0)
   {
      // then don't allow this plugin to run
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (bad dirname: '%s')", Plugin_info.name, dirname);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (bad dirname: '%s')", Plugin_info.name, dirname);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }

   // print a message to notify about plugin attaching
   LOG_CONSOLE (PLID, "%s: plugin attaching", Plugin_info.name);
   LOG_MESSAGE (PLID, "%s: plugin attaching", Plugin_info.name);

   // ask the engine to register the CVARs this plugin uses
   CVAR_REGISTER (&maxhumans);
   CVAR_REGISTER (&allowadminson);
   CVAR_REGISTER (&addbot_command);

   return (TRUE); // returning TRUE enables metamod to attach this plugin
}


C_DLLEXPORT int Meta_Detach (PLUG_LOADTIME now, PL_UNLOAD_REASON reason)
{
   // this function is called when metamod unloads the plugin. A basic check is made in order
   // to prevent unloading the plugin if its processing should not be interrupted.

   // is metamod allowed to unload the plugin ?
   if ((now > Plugin_info.unloadable) && (reason != PNL_CMD_FORCED))
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from unloading this plugin
   }

   return (TRUE); // returning TRUE enables metamod to unload this plugin
}


C_DLLEXPORT int GetEntityAPI2 (DLL_FUNCTIONS *pFunctionTable, int *interfaceVersion)
{
   // this is one of the initialization functions hooked by metamod in the gamedll API

   // keep track of pointers to the functions we use in the gamedll API
   gFunctionTable.pfnClientConnect = ClientConnect;
   gFunctionTable.pfnClientDisconnect = ClientDisconnect;
   gFunctionTable.pfnStartFrame = StartFrame;

   // copy the whole table for metamod to know which functions we are using here
   memcpy (pFunctionTable, &gFunctionTable, sizeof (DLL_FUNCTIONS));
   return (TRUE); // alright
}


C_DLLEXPORT int GetEngineFunctions (enginefuncs_t *pengfuncsFromEngine, int *interfaceVersion)
{
   // this is one of the initialization functions hooked by metamod in the engine API

   // keep track of pointers to the functions we use in the engine API
   meta_engfuncs.pfnCreateFakeClient = CreateFakeClient;

   // copy the whole table for metamod to know which functions we are using here
   memcpy (pengfuncsFromEngine, &meta_engfuncs, sizeof (enginefuncs_t));
   return (TRUE); // alright
}


// END of Metamod stuff




#ifdef _WIN32
int WINAPI DllMain (HINSTANCE hinstDLL, unsigned long fdwReason, LPVOID lpvReserved)
{
   // required Win32 DLL entry point. Good place for freeing stuff upon DLL unloading (in case
   // fdwReason would be DLL_PROCESS_DETACH) if needed. The same for Linux would be implemented
   // in an optional _fini() function.

   return (TRUE); // nothing to do here
}
#endif


void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals)
{
   // this function is called by the engine to tell the MOD DLL (hence the plugins too) the
   // address of its function table and its server global variables structure.

   // keep track of the engine functions and the server globals for our own use
   memcpy (&g_engfuncs, pengfuncsFromEngine, sizeof (enginefuncs_t));
   gpGlobals = pGlobals;
}


BOOL ClientConnect (edict_t *pEntity, const char *pszName, const char *pszAddress, char szRejectReason[128])
{
   // this function is called in order to tell the MOD DLL that a client attempts to connect the
   // game. The entity pointer of this client is pEntity, the name under which he connects is
   // pointed to by the pszName pointer, and its IP address string is pointed by the pszAddress
   // one. Note that this does not mean this client will actually join the game ; he could as
   // well be refused connection by the server later, because of latency timeout, unavailable
   // game resources, or whatever reason. In which case the reason why the game DLL (read well,
   // the game DLL, *NOT* the engine) refuses this player to connect will be printed in the
   // szRejectReason string in all letters. Understand that a client connecting process is done
   // in three steps. First, the client requests a connection to the server. This is engine
   // internals. When there are already too many players, the engine will refuse this client to
   // connect, and the game DLL won't even notice. Second, if the engine sees no problem, the
   // game DLL is asked. This is where we are. Once the game DLL acknowledges the connection,
   // the client downloads the resources it needs and synchronizes its local engine with the one
   // of the server. And then, the third step, which comes *AFTER* ClientConnect(), is when the
   // client officially enters the game, through the ClientPutInServer() function, later below.
   // Here we hook this function in order to keep the max humans count in the specified limits.
   // Also we check whether a bot needs to be disconnected, in order to always have one free
   // slot on the server for incoming clients. One final note, we can only hook player
   // connections, because bots call directly the right function in the GameDLL, without passing
   // through the metamod API. Remember the infrastructure:
   // Engine DLL <===> metamod + plugin <===> bot DLL <===> Game DLL

   int client_index;
   edict_t *pClient, *pBotClient;
   static char command[256];

   // the client is necessarily a human player, as explained above

   // is this player an administrator AND are admins allowed to enforce their connection ?
   if (IsAdministrator (pEntity) && (CVAR_GET_FLOAT ("allowadminson") > 0))
   {
      // this player is an administrator

      // are we reaching the max players count with this client ?
      if (player_count + 1 == gpGlobals->maxClients)
      {
         // cycle through all players
         for (client_index = 1; client_index <= gpGlobals->maxClients; client_index++)
         {
            pClient = INDEXENT (client_index); // get a pointer to this client's entity

            if (FNullEnt (pClient) || pClient->free || !(pClient->v.flags & (FL_CLIENT | FL_FAKECLIENT)))
               continue; // skip invalid players

            // is this client a bot ?
            if (pClient->v.flags & FL_FAKECLIENT)
               pBotClient = pClient; // remember the last bot we saw
         }

         // now have we found a bot ? if there were several bots, kick the last one
         if (!FNullEnt (pBotClient))
         {
            sprintf (command, "kick \"%s\"\n", STRING (pBotClient->v.netname));
            SERVER_COMMAND (command); // kick this bot
         }

         // else no bot was found, so notify the admin that we couldn't kick any bot
         else
            CLIENT_PRINTF (pEntity, print_console, MESSAGE_WARNING_COULDNOTKICKBOT); // notify admin
      }
   }
   else
   {
      // this player is not allowed to enforce his connection

      // are we reaching the max players count OR the max humans count with client ?
      if ((player_count + 1 == gpGlobals->maxClients)
          || (human_count == CVAR_GET_FLOAT ("maxhumans")))
      {
         // don't allow this human to connect
         sprintf (szRejectReason, MESSAGE_REFUSE_CLIENTCONNECT);
         RETURN_META_VALUE (MRES_SUPERCEDE, FALSE); // don't call ClientConnect in the game DLL
      }
   }

   RETURN_META_VALUE (MRES_IGNORED, 0);
}


void ClientDisconnect (edict_t *pEntity)
{
   // this function is called whenever a client is VOLUNTARILY disconnected from the server,
   // either because the client dropped the connection, or because the server dropped him from
   // the game (latency timeout). The effect is the freeing of a client slot on the server. Note
   // that clients and bots disconnected because of a level change NOT NECESSARILY call this
   // function, because in case of a level change, it's a server shutdown, and not a normal
   // disconnection. Note that we can only hook player disconnections, because bots call
   // directly the right function in the GameDLL, without passing through the metamod API,
   // *UNLESS* the disconnection is provoked by the engine, either because of a mapchange, or
   // because of a server command ; thus we need to ensure the entity disconnecting is really
   // a human player. Remember the infrastructure:
   // Engine DLL <===> metamod + plugin <===> bot DLL <===> Game DLL

   // don't care about bots
   if (pEntity->v.flags & FL_FAKECLIENT)
      RETURN_META (MRES_IGNORED);

   // are administrators allowed to enforce their connection
   // AND was this player an administrator ?
   if ((CVAR_GET_FLOAT ("allowadminson") > 0) && IsAdministrator (pEntity))
   {
      // were we at the max players count with this client ?
      if (player_count + 1 == gpGlobals->maxClients)
         add_bot_now = TRUE; // then we need to add a bot
   }

   RETURN_META (MRES_IGNORED);
}


void StartFrame (void)
{
   // this function starts a video frame. It is called once per video frame by the engine. If
   // you run Half-Life at 90 fps, this function will then be called 90 times per second. By
   // placing a hook on it, we have a good place to do things that should be done continuously
   // during the game, for example we have to control the bot population, so that if a new
   // player joins the server, a bot should be kicked, and if the player population decreases,
   // other bots should join, because always keeping an open slot will take the player
   // connecting approval from the engine level to the MOD DLL level, which is what we need
   // because we have to ensure certain players will always be allowed to connect. Local
   // variables have been declared static to speedup the function execution a bit.

   static int client_index;
   static edict_t *pPlayer;
   static char server_command[256];

   bot_count = 0; // prepare for counting the number of bots
   human_count = 0; // prepare for counting the number of humans

   // cycle through all players
   for (client_index = 1; client_index <= gpGlobals->maxClients; client_index++)
   {
      pPlayer = INDEXENT (client_index); // get a pointer to this player's entity

      if (FNullEnt (pPlayer) || pPlayer->free || !(pPlayer->v.flags & (FL_CLIENT | FL_FAKECLIENT)))
         continue; // skip invalid players

      // is this player a bot or a human ?
      if (pPlayer->v.flags & FL_FAKECLIENT)
         bot_count++; // this player is a bot, increase the bot count
      else
         human_count++; // this player is a human client, increase the human count
   }

   // now that the loop has passed, we know the number of bots and the number of humans
   player_count = human_count + bot_count; // compute the total number of players on this server

   // are we supposed to add a bot now because an admin disconnected ?
   if (add_bot_now)
   {
      sprintf (server_command, CVAR_GET_STRING ("addbot_command")); // get the addbot command syntax

      // Yes, didn't you know that "pb" prefix was in fact a CVAR ?
      CVAR_SET_STRING ("pb", server_command); // server addbot command syntax is 'pb addbot'
      add_bot_now = FALSE; // reset the flag not to do it twice
   }

   RETURN_META (MRES_IGNORED);
}


edict_t *CreateFakeClient (const char *netname)
{
   // this function is called when the bot DLL asks the engine to create a bot. Since we don't
   // want bots to be created when the max player count is reached, so as to always let one slot
   // open on the server, we either allow, or disallow, the engine function to be called.

   // has the max player count already been reached ?
   if (player_count + 1 == gpGlobals->maxClients)
      RETURN_META_VALUE (MRES_SUPERCEDE, NULL); // don't allow the bot to be created

   RETURN_META_VALUE (MRES_IGNORED, NULL); // else allow the plugin to call the function normally
}


bool IsAdministrator (edict_t *pPlayer)
{
   // this function returns TRUE if the WONId of the player whose player entity is pointed to by
   // pPlayer is present in the FILE_ADMINPLAYERS file (currently "cstrike/serveradmins.cfg"),
   // which means this player has the right to enforce a connection to this server. In order
   // to check this, we get the player's WONId by asking the engine, and compare the returned
   // value to each of the lines of the file. If a match is found, this player is assumed to be
   // an administrator and we return TRUE; else, if no match is found, we return FALSE.

   FILE *fp;
   char wonid[32], line_buffer[256];
   bool match = FALSE;

   if (FNullEnt (pPlayer))
      return (FALSE); // reliability check

   // open the admin players file in read only mode
   fp = fopen (FILE_ADMINPLAYERS, "r"); // open the file for reading
   if (fp == NULL)
      return (FALSE); // if the file doesn't exist, this player can't be an administrator, eh :)

   // ask the engine for the WONId of the player and convert it to a string
   sprintf (wonid, "%d\n", g_engfuncs.pfnGetPlayerWONId (pPlayer));

   // now read the file line per line
   while (fgets (line_buffer, 255, fp) != NULL)
   {
      if ((line_buffer[0] == 0) || (line_buffer[0] == '\n') || (line_buffer[0] == '#') || (line_buffer[0] == ';'))
         continue; // ignore line if void or commented

      // now compare the recorded WONId to the one of the player
      if (strcmp (wonid, line_buffer) == 0)
      {
         fclose (fp); // a match was found, close the file
         return (TRUE); // this player is an administrator
      }
   }

   // all lines were parsed, but we didn't found a match

   fclose (fp); // don't forget to close the file
   return (FALSE); // player is not an administrator, return FALSE
}


void GetGameDir (char *game_dir)
{
   // This function fixes the erratic behaviour caused by the use of the GET_GAME_DIR engine
   // macro, which returns either an absolute directory path, or a relative one, depending on
   // whether the game server is run standalone or not. This one always return a RELATIVE path.

   unsigned char length, fieldstart, fieldstop;

   GET_GAME_DIR (game_dir); // call the engine macro and let it mallocate for the char pointer

   length = strlen (game_dir); // get the length of the returned string
   length--; // ignore the trailing string terminator

   // format the returned string to get the last directory name
   fieldstop = length;
   while (((game_dir[fieldstop] == '\\') || (game_dir[fieldstop] == '/')) && (fieldstop > 0))
      fieldstop--; // shift back any trailing separator

   fieldstart = fieldstop;
   while ((game_dir[fieldstart] != '\\') && (game_dir[fieldstart] != '/') && (fieldstart > 0))
      fieldstart--; // shift back to the start of the last subdirectory name

   if ((game_dir[fieldstart] == '\\') || (game_dir[fieldstart] == '/'))
      fieldstart++; // if we reached a separator, step over it

   // now copy the formatted string back onto itself character per character
   for (length = fieldstart; length <= fieldstop; length++)
      game_dir[length - fieldstart] = game_dir[length];
   game_dir[length - fieldstart] = 0; // terminate the string

   return;
}
